-- Puffin: A Tojaso FowlPlay Project
-- Model component for tracking player-specific information

local Model_Player = {}
local EF = Puffin.ErrorFormat
local DB, UN
local stance_changed = true
local weapon_changed = true
local power_changed = true
local talent_changed = true
local specialization_changed = true
local aura_changed = true
local bufftooltip = nil -- used to store tooltip for scanning weapon buffs
local bufftooltipline = {} -- caches lines in weapon tooltips
local mhLastBuff = nil -- saves name of most recent main hand weapon buff
local ohLastBuff = nil -- saves name of most recent off hand weapon buff
local weaponBuffDurations = {} -- cache of previously seen weapon buffs
local dispelTypes = {} -- table of debuff types that the character can dispel
local stanceClasses = { WARRIOR = true, PALADIN = true, MONK = true } -- classes with stances to add to player buffs

local function WeaponEvent() weapon_changed = true end
local function StanceEvent() stance_changed = true end
local function PowerEvent(e, unit) if unit == "player" then power_changed = true end end
local function TalentEvent() talent_changed = true; aura_changed = true end
local function AuraEvent(e, unit) if unit == "player" then aura_changed = true end end

-- Initialize tooltip to be used for determining weapon buffs
-- This code is based on the Pitbull implementation
local function InitializeBuffTooltip()
	bufftooltip = CreateFrame("GameTooltip", nil, UIParent)
	bufftooltip:SetOwner(UIParent, "ANCHOR_NONE")
	local fs = bufftooltip:CreateFontString()
	fs:SetFontObject(_G.GameFontNormal)
	for i = 1, 30 do
		local ls = bufftooltip:CreateFontString()
		ls:SetFontObject(_G.GameFontNormal)
		bufftooltip:AddFontStrings(ls, fs)
		bufftooltipline[i] = ls
	end
end

-- No easy way to get this info, so scan item slot info for mainhand and offhand weapons using a tooltip
-- Weapon buffs are usually formatted in tooltips as name strings followed by remaining time in parentheses
-- This routine scans the tooltip for the first line that is in this format and extracts the weapon buff name without rank or time
local function GetWeaponBuffName(weaponSlot)
	bufftooltip:ClearLines()
	if not bufftooltip:IsOwned(UIParent) then bufftooltip:SetOwner(UIParent, "ANCHOR_NONE") end
	bufftooltip:SetInventoryItem("player", weaponSlot)
	for i = 1, 30 do
		local text = bufftooltipline[i]:GetText()
		if text then
			local name = text:match("^(.+) %(%d+ [^$)]+%)$") -- extract up to left paren if match weapon buff format
			if name then
				name = (name:match("^(.*) %d+$")) or name -- remove any trailing numbers
				if shamanEnchants then -- special case for localing shaman weapon enhancements
					for _, enchant in pairs(shamanEnchants) do if string.find(enchant, name) then name = enchant; break end end
				end
				return name
			end
		else
			break
		end
	end
	return nil
end

-- Initialize the dispel table which lists what types of debuffs the player can dispel
-- This needs to be updated when the player changes talent specs or learns new spells
local function SetDispelDefaults()
	local myClass = Puffin.myClass
	dispelTypes.Poison = false; dispelTypes.Curse = false; dispelTypes.Magic = false; dispelTypes.Disease = false
	if myClass == "DRUID" then
		if IsSpellKnown(88423) then -- Nature's Cure
			dispelTypes.Poison = true; dispelTypes.Curse = true; dispelTypes.Magic = true
		elseif IsSpellKnown(2782) then -- Remove Corruption
			dispelTypes.Poison = true; dispelTypes.Curse = true
		end
	elseif myClass == "MONK" then
		if IsSpellKnown(115451) then -- Internal Medicine
			dispelTypes.Poison = true; dispelTypes.Disease = true; dispelTypes.Magic = true
		elseif IsSpellKnown(115450) then -- Detox
			dispelTypes.Poison = true; dispelTypes.Disease = true
		end
	elseif myClass == "PRIEST" then
		if IsSpellKnown(527) then
			dispelTypes.Magic = true; dispelTypes.Disease = true -- Purify
		elseif IsSpellKnown(32375) then
			dispelTypes.Magic = true -- Mass Dispel
		end
	elseif myClass == "MAGE" then
		if IsSpellKnown(475) then dispelTypes.Curse = true end -- Remove Curse
	elseif myClass == "PALADIN" then
		if IsSpellKnown(4987) then -- Cleanse
			dispelTypes.Poison = true; dispelTypes.Disease = true
			if IsSpellKnown(53551) then dispelTypes.Magic = true end -- Sacred Cleansing
		end
	elseif myClass == "SHAMAN" then
		if IsSpellKnown(77130) then
			dispelTypes.Curse = true; dispelTypes.Magic = true -- Purify Spirit
		elseif IsSpellKnown(51886) then
			dispelTypes.Curse = true -- Cleanse Spirit
		end
	end
	updateDispels = false
end

-- Return true if the player can dispel the type of debuff on the unit
local function IsDebuffDispellable(unit, debuffType)
	if not debuffType then return false end
	if updateDispels == true then MOD:SetDispelDefaults() end
	local t = dispelTypes[debuffType]
	if not t then return false end
	if (t == "player") and (unit ~= "player") then return false end -- special case for self-only dispels
	if unit == "player" then return true end -- always can dispel debuffs on self
	if UnitIsFriend("player", unit) ~= nil then return true end -- only can dispel on friendly units
	return false
end

function Model_Player.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, "Puffin.Model_Player: invalid reference to data management component")
	DB.AllocateDatabase("player", false)
	UN = Puffin.Component("Model_Units")
	assert(UN, EF("Adapter_SpellTrackers", "Model_Units component not found"))
	InitializeBuffTooltip()

	Puffin.RegisterCallback("UNIT_INVENTORY_CHANGED", WeaponEvent)
	Puffin.RegisterCallback("UPDATE_SHAPESHIFT_FORM", StanceEvent)
	Puffin.RegisterCallback("UNIT_POWER", PowerEvent)
	Puffin.RegisterCallback("PLAYER_TALENT_UPDATE", TalentEvent)
	Puffin.RegisterCallback("PLAYER_SPECIALIZATION_CHANGED", TalentEvent)	
	Puffin.RegisterCallback("SPELLS_CHANGED", TalentEvent)
	Puffin.RegisterCallback("UNIT_AURA", AuraEvent)
end

function Model_Player.OnUpdate()
	local changed = false
	if stance_changed and stanceClasses[Puffin.myClass] then
		stance_changed = false; changed = true
		DB.SetAttribute("player", "stance", "name", nil)
		local stance = GetShapeshiftForm()
		if stance and stance > 0 then
			local _, name = GetShapeshiftFormInfo(stance)
			if name then
				local icon = GetSpellTexture(name)
				local link = GetSpellLink(name)
				DB.SetAttribute("player", "stance", "name", name)
				DB.SetAttribute("player", "stance", "icon", icon)
				DB.SetAttribute("player", "stance", "link", link)
			end
		end
	end

	if weapon_changed then
		weapon_changed = false; changed = true
		DB.SetAttribute("player", "mainhand", "name", nil)
		DB.SetAttribute("player", "offhand", "name", nil)
		-- first check if there are weapon auras then, only if necessary, use tooltip to scan for the buff names
		local mh, mhms, mhc, oh, ohms, ohc = GetWeaponEnchantInfo()
		if mh then -- add the mainhand buff, if any, to the table
			local islot = GetInventorySlotInfo("MainHandSlot")
			local mhbuff = GetWeaponBuffName(islot)
			if not mhbuff then -- if tooltip scan fails then use fallback of weapon name or slot name
				local weaponLink = GetInventoryItemLink("player", islot)
				if weaponLink then mhbuff = GetItemInfo(weaponLink) end
				if not mhbuff then mhbuff = "Mainhand" end
			end
			local icon = GetInventoryItemTexture("player", islot)					
			local expire = GetTime() + (mhms / 1000)
			DB.SetAttribute("player", "mainhand", "name", mhbuff .. " Weapon")
			DB.SetAttribute("player", "mainhand", "icon", icon)
			DB.SetAttribute("player", "mainhand", "count", mhc)
			DB.SetAttribute("player", "mainhand", "expire", expire)
			mhLastBuff = mhbuff -- caches the name of the weapon buff so can clear it later
		else mhLastBuff = nil end
		
		if oh then -- add the offhand buff, if any, to the table
			local islot = GetInventorySlotInfo("SecondaryHandSlot")
			local ohbuff = GetWeaponBuffName(islot)
			if not ohbuff then -- if tooltip scan fails then use fallback of weapon name or slot name
				local weaponLink = GetInventoryItemLink("player", islot)
				if weaponLink then ohbuff = GetItemInfo(weaponLink) end
				if not ohbuff then ohbuff = "Offhand" end
			end
			local icon = GetInventoryItemTexture("player", islot)
			local expire = GetTime() + (ohms / 1000)
			DB.SetAttribute("player", "offhand", "name", ohbuff .. " Weapon")
			DB.SetAttribute("player", "offhand", "icon", icon)
			DB.SetAttribute("player", "offhand", "count", ohc)
			DB.SetAttribute("player", "offhand", "expire", expire)
			ohLastBuff = ohbuff -- caches the name of the weapon buff so can clear it later
		else ohLastBuff = nil end
	end
	
	if power_changed then
		power_changed = false; changed = true
		DB.SetAttribute("player", "power", "name", nil)
		local power, id = nil, nil
		if Puffin.myClass == "PALADIN" then power = UnitPower("player", SPELL_POWER_HOLY_POWER); id = 85247
		elseif Puffin.myClass == "PRIEST" then power = UnitPower("player", SPELL_POWER_SHADOW_ORBS); id = 95740
		elseif Puffin.myClass == "MONK" then power = UnitPower("player", SPELL_POWER_LIGHT_FORCE); id = 97272
		elseif Puffin.myClass == "WARLOCK" then
			if IsSpellKnown(108647) then
				power = UnitPower("player", SPELL_POWER_BURNING_EMBERS, true); id = 108647
			elseif IsSpellKnown(1120) then
				power = UnitPower("player", SPELL_POWER_SOUL_SHARDS); id = 117198
			elseif IsSpellKnown(104315) then
				power = UnitPower("player", SPELL_POWER_DEMONIC_FURY); id = 104315
			end
		end
		if power and power > 0 then
			local name, _, icon = GetSpellInfo(id)
			local link = GetSpellLink(id)
			if name then
				DB.SetAttribute("player", "power", "name", name)
				DB.SetAttribute("player", "power", "icon", icon)
				DB.SetAttribute("player", "power", "link", link)
				DB.SetAttribute("player", "power", "count", power)
				DB.SetAttribute("player", "power", "id", id)
			end
		end
	end
	
	if talent_changed then
		talent_changed = false; changed = true
		SetDispelDefaults()
		DB.SetAttribute("player", "dispel", "function", IsDebuffDispellable)
		local tabs = GetNumSpecializations(false, false)
		if tabs >= 0 then
			local currentSpec = GetSpecialization()
			local talentSpec = currentSpec and select(2, GetSpecializationInfo(currentSpec)) or "None"
			DB.SetAttribute("player", "specialization", "name", talentSpec)		
		
			local ts = GetNumTalents(currentSpec)
			for i = 1, ts do
				local name, _, _, _, selected = GetTalentInfo(i) -- player's active talents
				if name then
					DB.SetAttribute("player", "talent", name, selected)
				end
			end
		end
	end
	
	if aura_changed then
		aura_changed = false; changed = true
		DB.ReleaseRecord("player", "missing")
		local classSpells = DB.GetRecord("spells", "class")
		if classSpells then
			local cs = classSpells[Puffin.myClass]
			for id, t in pairs(cs) do
				if id ~= "exclusive" and IsSpellKnown(id) then -- check to make sure know the spell
					local spell, _, icon = GetSpellInfo(id)
					if spell then
						local found = false
						for _, buff in pairs(t) do -- then check if any of the equivalent buffs are active
							local name = GetSpellInfo(buff)
							if name then
								if UnitBuff("player", name) then found = true end
								if DB.GetAttribute("player", "stance", "name") == name then found = true end
								if DB.GetAttribute("player", "mainhand", "name") == name then found = true end
								if DB.GetAttribute("player", "offhand", "name") == name then found = true end
							end
						end
						if not found and cs.exclusive then -- make sure not exclusive with another spell that was cast by player
							local exspell, excluded = false, false
							for _, exid in pairs(cs.exclusive) do
								if id == exid then exspell = true else
									local exname = GetSpellInfo(exid)
									if exname and UnitBuff("player", exname, nil, "PLAYER") then excluded = true end
								end
							end
							if exspell and excluded then found = true end
						end
						if not found then DB.SetAttribute("player", "missing", spell, icon) end
					end
				end
			end
		end
	end
	
	if changed then Puffin.Callback(UN.callbacks["player"], "player", "changed") end
end

Puffin.RegisterComponent("Model_Player", Model_Player, "MODEL", 1)
