-- Puffin: A Tojaso FowlPlay Project
-- View component for an icon group that shows icons for all spells in a list.
-- Each icon is assembled from a collection of widgets, as defined by a layout.
-- An anchor frame is constructed and icons are displayed relative to it.

-- View API:
-- View_IconGroup.Create() - a new instance of an icon group view
-- View_IconGroup:Release() - release an instance of an icon group view, recycling it for future use
-- View_IconGroup:Configure() - configure based on current defaults and widget descriptions
-- View_IconGroup:Update() - update the icon group

-- View attributes:
-- offsetX - horizontal offset from UIParent [c]
-- offsetY - vertical offset from UIParent [c]
-- anchorFrame - frame to anchor the icon group to (default: UIParent) [c]
-- anchorPoint - if anchor defined then point to position relative to (default: BOTTOMLEFT) [c]
-- relativePoint - point on this widget to position relative to (default: BOTTOMLEFT) [c]
-- layout - table of attributes used to configure the icon group
-- sfx - animations, color changes, start and expiration sounds
-- depth - display depth for this view [c]
-- scale - factor to scale view's display size [d]
-- alpha - alpha for the icon group [d]
-- hidden - hide the view [d]

-- Layout attributes:
-- defaults - table with default settings, including those for each class of widget [c]
-- button - widget template for an action button [c]
-- timer - widget template for timer text field [c]
-- count - widget template for count text field [c]
-- backdrop - widget template for backdrop panel that dynamically scales with active icons [c]
-- orientation - "horizontal" or "vertical" (default: "horizontal") [c]
-- direction - true for "left" or "up", false for "right" or "down" [c]
-- limit - maximum number of icons to show (default: 10) [c]
-- spacingX - horizontal distance between icons (default: 0) [c]
-- spacingY - vertical distance between icons (default: 0) [c]
-- wrap - how many icons to show before wrapping to next row or column (default: 0, i.e., don't wrap) [c]
-- wrapDirection - true for "left" or "up", false for "right" or "down" [c]
-- center - if true then snap visible icons to center at anchor [d]

local View_IconGroup = {}
local EF = Puffin.ErrorFormat
local DB, VC
local classDepth = 0 -- mininum depth for view's of this class
local timelines = {} -- active timeline views
local cc = { a = 1 } -- temporary color table with alpha preset

-- Return an initialized instance of the view
function View_IconGroup.Create() return VC.CreateView() end

-- Release a view when no longer needed, recycle after clearing attributes and hiding widgets associated with it
function View_IconGroup:Release() timelines[self.name] = nil; VC.ReleaseView(self) end

-- Calculate the offset for a time value on a timeline
local function Timeline_Offset(tl, value)
	if value >= tl.duration then return tl.width end
	if value <= 0 then return 0 end
	return tl.width * ((value / tl.duration) ^ (1 / tl.exponent))
end

-- Configure all icons (with button, timer text, count text widgets) associated with this view
function View_IconGroup:Configure()
	assert(self.defaults, EF("View_IconGroup", "undefined defaults"))
	assert(self.layout, EF("View_IconGroup", "undefined layout"))
	local offsetDepth = ((self.depth or 0) + classDepth) * 10
	local xname = string.gsub(self.name, " ", "_")
	self.frame = CreateFrame("Frame", "PuffinIconGroup" .. xname, UIParent) -- create an anchor frame
	self.frame:SetFrameLevel(self.frame:GetFrameLevel() + 20) -- set frame level higher than other addons
	
	self.frame:SetSize(1, 1) -- one pixel size for the anchor frame
	self.frame:SetPoint(self.anchorPoint or "BOTTOMLEFT", self.anchorFrame or UIParent, self.relativePoint or "BOTTOMLEFT",
		self.offsetX or 0, self.offsetY or 0)
	self.frame:Show()

	-- create and configure the widgets for the maximum number of icons in the icon group
	local layout = self.layout
	local tl = layout.timeline
	assert(layout.button, EF("View_IconGroup", "incomplete layout")) -- required fields
	self.icons = Puffin.AllocateTable()
	for i = 1, (layout.limit or 10) do
		local icon = Puffin.AllocateTable()
		local t = VC.CreateWidget(layout.button)
		if t then
			icon.button = t; t.anchor = self.frame; t.parent = self.frame; t.masque = self.name
			t.onEnter = self.onEnter; t.onLeave = self.onLeave; t.onClick = self.onClick
			VC.ConfigureWidget(self, t, layout.button, offsetDepth)
		end
		if layout.timer then
			t = VC.CreateWidget(layout.timer)
			if t then icon.timer = t; t.anchor = icon.button; t.parent = icon.button; VC.ConfigureWidget(self, t, layout.timer, offsetDepth) end
		end
		if layout.count then
			t = VC.CreateWidget(layout.count)
			if t then icon.count = t; t.anchor = icon.button; t.parent = icon.button; VC.ConfigureWidget(self, t, layout.count, offsetDepth) end
		end
		self.icons[i] = icon
	end

	if layout.backdrop then -- create display panel widget for optional backdrop
		local t = VC.CreateBackdrop(layout.backdrop); self.backdrop = t
		if tl then -- add font strings to timeline backdrop and configure (non-timeline backdrops defer configuration)
			t.anchor = self.frame; t.anchorPoint = "TOPRIGHT"; t.relativePoint = "TOPRIGHT"
			VC.ConfigureBackdrop(self, t, layout.backdrop, offsetDepth)
			if tl.caption and tl.captions then
				local isVertical = (layout.orientation == "vertical")
				local dir = layout.direction and 1 or -1 -- plus or minus depending on direction
				local w, h, edge, x1, y1
				if isVertical then
					w = tl.height; h = tl.width; edge = layout.direction and "TOP" or "BOTTOM"
				else
					w = tl.width; h = tl.height; edge = layout.direction and "RIGHT" or "LEFT"
				end
				local i = 0
				self.captions = Puffin.AllocateTable()
				for caption, offset in pairs(tl.captions) do
					local cap = VC.CreateWidget(tl.caption)
					i = i + 1; self.captions[i] = cap; cap.text = caption; cap.parent = t; cap.hidden = false
					local delta = Timeline_Offset(tl, offset)
					if isVertical then x1 = 0; y1 = (delta * dir) - h else x1 = (delta * dir) - w; y1 = 0 end
					cap.anchorPoint = edge; cap.relativePoint = edge; cap.offsetX = x1; cap.offsetY = y1; cap.anchor = t
					VC.ConfigureWidget(self, cap, tl.caption, offsetDepth)
					cap._class.Update(cap)
				end
			end
		end
	end
	if tl then timelines[self.name] = self end
end

-- Set the depth for an icon's widgets
local function SetIconDepth(icon, layout, depth)
	local t = icon.button; t.depth = depth + (layout.button.depth or 0); t._class.SetDepth(t, t.depth)
	local t = icon.timer; if layout.timer and t then t.depth = depth + (layout.timer.depth or 0); t._class.SetDepth(t, t.depth) end
	local t = icon.count; if layout.count and t then t.depth = depth + (layout.count.depth or 0); t._class.SetDepth(t, t.depth) end
end

-- Update icon positions on a timeline
local function UpdateTimeline(ig, layout, spellCount, spellList, anchorFrame)
	if not spellCount or not spellList or not anchorFrame then return end
	ig.count = 0 -- cache for animation
	local dir = layout.direction and 1 or -1 -- plus or minus depending on direction
	local isVertical = (layout.orientation == "vertical")
	local level, ascend, z = anchorFrame:GetFrameLevel() + 10, (math.floor(GetTime()) % 4) >= 2, 0
	local w, h, edge, lastDelta, x1, y1
	local tl = layout.timeline
	local fudge = tl.overlapOffset or 0
	assert(tl.width and tl.height and tl.exponent and tl.duration, EF("View_IconGroup", "incomplete timeline layout")) -- required fields
	if isVertical then
		w = tl.height; h = tl.width; edge = layout.direction and "TOP" or "BOTTOM"
	else
		w = tl.width; h = tl.height; edge = layout.direction and "RIGHT" or "LEFT"
	end
	for i = 1, layout.limit do
		local icon = ig.icons[i]
		if i <= spellCount then
			local s = spellList[i]
			local delta = Timeline_Offset(tl, s.expire - GetTime())
			if i > 1 and math.abs(delta - lastDelta) < math.abs(fudge) then z = z + 1; delta = lastDelta + (fudge * z) else z = 0; lastDelta = delta end
			if isVertical then x1 = (w + 1) / -2; y1 = (delta * dir) - h else x1 = (delta * dir) - w; y1 = (h + 1) / -2 end
			SetIconDepth(icon, layout, ascend and (level + ((spellCount - i) * 10)) or (level + (i * 10)))
			local t = icon.button; t._class.SetPosition(t, edge, anchorFrame, edge, math.floor((x1 * 4) + 0.5) / 4, math.floor((y1 * 4) + 0.5) / 4)
			if ig.sfx then
				if ig.sfx.clockOverlay and s.duration > 0 then
					t.clockExpire = s.expire; t.clockDuration = s.duration; t.clockDirection = ig.sfx.clockDirection
				else
					t.clockDuration = nil
				end
				if ig.sfx.masqueColor then cc.r = s.r; cc.g = s.g; cc.b = s.b; t.masqueColor = cc else t.masqueColor = nil end
			end
			local e = t.eventHandler; if e then e.source = s.source; e.tooltip = s.tooltip; e.unit = s.unit; e.caster = s.caster end
			t.hidden = false; t.texture = s.icon; t._expire = s.expire; t._class.Update(t)
			t = icon.timer; if t then
				t.hidden = false; t.timeMax = s.duration; t.timeExpire = s.expire; t._class.Update(t)
			end
			t = icon.count; if t then
				if s.forceCount or (s.count or 0) > 1 then t.hidden = false; t.text = tostring(s.count or 0) else t.hidden = true; t.text = "" end; t._class.Update(t)
			end
			ig.count = ig.count + 1
		else
			local t = icon.button; t.hidden = true; table.wipe(t.eventHandler); t._class.Update(t)
			t = icon.timer; if t then t.hidden = true; t._class.Update(t) end
			t = icon.count; if t then t.hidden = true; t._class.Update(t) end
		end
	end
end

-- Animate icon positions on a timeline
local function RefreshTimeline(ig)
	local layout, anchorFrame, iconCount = ig.layout, ig.frame, ig.count
	if not layout or not iconCount or not anchorFrame then return end
	local dir = layout.direction and 1 or -1 -- plus or minus depending on direction
	local isVertical = (layout.orientation == "vertical")
	local level, ascend, z = anchorFrame:GetFrameLevel() + 10, (math.floor(GetTime()) % 4) >= 2, 0
	local w, h, edge, lastDelta
	local tl = layout.timeline
	local fudge = tl.overlapOffset or 0
	assert(tl.width and tl.height and tl.exponent and tl.duration, EF("View_IconGroup", "incomplete timeline layout")) -- required fields
	if isVertical then
		w = tl.height; h = tl.width; edge = layout.direction and "TOP" or "BOTTOM"
	else
		w = tl.width; h = tl.height; edge = layout.direction and "RIGHT" or "LEFT"
	end
	for i = 1, iconCount do
		local icon = ig.icons[i]
		local t = icon.button
		local delta = Timeline_Offset(tl, t._expire - GetTime())
		if i > 1 and math.abs(delta - lastDelta) < math.abs(fudge) then z = z + 1; delta = lastDelta + (fudge * z) else z = 0; lastDelta = delta end
		if isVertical then x1 = (w + 1) / -2; y1 = (delta * dir) - h else x1 = (delta * dir) - w; y1 = (h + 1) / -2 end
		SetIconDepth(icon, layout, ascend and (level + ((iconCount - i) * 10)) or (level + (i * 10)))
		local t = icon.button; t._class.SetPosition(t, edge, anchorFrame, edge, math.floor((x1 * 4) + 0.5) / 4, math.floor((y1 * 4) + 0.5) / 4)
	end
end
	
-- Update all icons associated with this view based on spell list
function View_IconGroup:Update(spellCount, spellList)
	local layout = self.layout
	assert(layout, EF("View_IconGroup", "undefined layout"))
	local wrap = 0 -- indicates default of not wrapping
	local dir = layout.direction and 1 or -1 -- will be 1 or -1 depending on growth direction
	local x0, y0 = 0, 0 -- starting position is offset if snapping to center
	local sx, sy = layout.spacingX or 0, layout.spacingY or 0
	local dx, dy = dir * (layout.width + sx), dir * (layout.height + sy) -- offset to each new icon
	local wx, wy = 0, 0 -- offsets from starting point when need to wrap
	local bw, bh = 0, 0 -- number of icon widths and heights in size of backdrop
	local xoffset, yoffset, xdir, ydir, wadjust = 0, 0, 1, dir, 0 -- position adjustments for backdrop
	local count, maxBars, cdir = spellCount, layout.limit, 0
	if not maxBars or (maxBars == 0) then maxBars = count end
	if count > maxBars then count = maxBars end
	local ac = count -- actual count before wrap adjustment
	if layout.wrap then wrap = layout.wrap; if (wrap > 0) and (count > wrap) then count = wrap end end
	local anchorPoint = "BOTTOMLEFT"
	if layout.orientation == "vertical" then
		wx = -dx; dx = 0; bh = count; if count > 0 then bw = math.ceil(ac / count) else bw = 1 end
		if not layout.direction then anchorPoint = "TOPLEFT"; wx = -wx; cdir = -1 end
		if not layout.wrapDirection then xoffset = -(bw - 1) * (layout.width + sx) end
		if layout.center then local z = (dy * (((count - dir) / 2) + cdir)); y0 = y0 - z; yoffset = yoffset - z end
		if layout.wrapDirection then wx = -wx end
		if wrap > 0 then -- attachment options differ when wrapping
			self.lastRow = x0 + (wx * bw); self.lastColumn = y0 + (dy * count); self.lastX = nil; self.lastY = nil
		else
			self.lastX = x0; self.lastY = y0 + (dy * count); self.lastRow = nil; self.lastColumn = nil
		end
	else -- horizontal
		wy = dy; dy = 0; bw = count; if count > 0 then bh = math.ceil(ac / count) else bh = 1 end
		if layout.direction then anchorPoint = "BOTTOMRIGHT"; wy = -wy; cdir = -1 end
		if not layout.wrapDirection then yoffset = -(bh - 1) * (layout.height + sy) end			
		if dir < 0 then
			xoffset = dir * (bw - 1) * (layout.width + sx); ydir = -ydir
		else
			xoffset = (bw - 1) * (layout.width + sx); xdir = -1			
		end
		if layout.center then local z = (dx * (((count + dir) / 2) + cdir)); x0 = x0 - z; xoffset = xoffset - z end
		if layout.wrapDirection then wy = -wy end
		if wrap > 0 then -- attachment options differ when wrapping
			self.lastRow = x0 + (dx * count); self.lastColumn = y0 + (wy * bh); self.lastX = nil; self.lastY = nil
		else
			self.lastX = x0 + (dx * count); self.lastY = y0; self.lastRow = nil; self.lastColumn = nil
		end
	end

	local anchorFrame = self.frame
	local tl = layout.timeline
	anchorFrame:SetAlpha(self.alpha or 1); anchorFrame:SetScale(self.scale or 1)
	if tl then -- update icon positions on timeline
		UpdateTimeline(self, layout, count, spellList, anchorFrame)
	else -- not timeline so less restricted layout options
		self.count = 0
		for i = 1, layout.limit do
			local icon = self.icons[i]
			if i <= spellCount then
				local w, skip = i - 1, 0; if wrap > 0 then skip = floor(w / wrap); w = w % wrap end
				local x1 = x0 + (dx * w) + (wx * skip); local y1 = y0 + (dy * w) + (wy * skip)
				local s = spellList[i]
				local t = icon.button
				self.count = self.count + 1
				t._class.SetPosition(t, anchorPoint, anchorFrame, anchorPoint, x1, y1)
				if self.sfx then
					if self.sfx.clockOverlay and s.duration > 0 then
						t.clockExpire = s.expire; t.clockDuration = s.duration; t.clockDirection = self.sfx.clockDirection
					else
						t.clockDuration = nil
					end
					if self.sfx.masqueColor then cc.r = s.r; cc.g = s.g; cc.b = s.b; t.masqueColor = cc else t.masqueColor = nil end
				end
				local e = t.eventHandler; if e then e.source = s.source; e.tooltip = s.tooltip; e.unit = s.unit; e.caster = s.caster end
				t.hidden = false; t.texture = s.icon; t._class.Update(t)
				t = icon.timer; if t then
					t.hidden = false; t.timeMax = s.duration; t.timeExpire = s.expire; t._class.Update(t)
				end
				t = icon.count; if t then
					if s.forceCount or (s.count or 0) > 1 then t.hidden = false; t.text = tostring(s.count or 0) else t.hidden = true; t.text = "" end; t._class.Update(t)
				end
			else
				local t = icon.button; t.hidden = true; table.wipe(t.eventHandler); t._class.Update(t)
				t = icon.timer; if t then t.hidden = true; t._class.Update(t) end
				t = icon.count; if t then t.hidden = true; t._class.Update(t) end
			end
		end
	end

	local t = self.backdrop -- optional backdrop adjusted after icons are arranged
	if t then
		if spellCount > 0 or (tl and not tl.hideEmpty) then
			t.hidden = false
			if not tl then
				local offsetDepth = ((self.depth or 0) + classDepth) * 10	
				t.width = bw * layout.width; if bw > 1 then t.width = t.width + ((bw - 1) * sx) end
				t.height = bh * layout.height; if bh > 1 then t.height = t.height + ((bh - 1) * sy) end
				t.anchor = anchorFrame; t.anchorPoint = anchorPoint; t.relativePoint = anchorPoint; t.offsetX = xoffset; t.offsetY = yoffset
				VC.ConfigureBackdrop(self, t, layout.backdrop, offsetDepth)
			end
		else
			t.hidden = true
		end
		t._class.Update(t)
	end
end

function View_IconGroup.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, EF("View_IconGroup", "Data component not found"))
	VC = Puffin.Component("View_Class") -- link to view class functions
	assert(VC, EF("View_IconGroup", "View class component not found"))
end

function View_IconGroup:OnRefresh() for _, ig in pairs(timelines) do RefreshTimeline(ig) end end

Puffin.RegisterComponent("View_IconGroup", View_IconGroup, "VIEW", 2)
