-- Puffin: A Tojaso FowlPlay Project
-- Component containing class functions shared by view components.

-- Exported functions:
-- CreateView() - return a new view (or reuse a recycled view)
-- ReleaseView() - release view-specific resources and recycle the view
-- CheckMedia(v) - check an attribute to see if it refers to special media
-- CreateWidget(w) - return a new instance of a widget based on a widget descriptor
-- ConfigureWidget(group, t, w, depth, noDimensions) - configure a widget using group and descriptor settings
-- CreateBackdrop(w) - return a backdrop based on a panel widget descriptor
-- ConfigureBackdrop(group, t, w, depth, noDimensions) - configure a backdrop for a group

-- View attributes:
-- offsetX - horizontal offset from UIParent [c]
-- offsetY - vertical offset from UIParent [c]
-- anchorFrame - frame to anchor the icon group to (default: UIParent) [c]
-- anchorPoint - if anchor defined then point to position relative to (default: BOTTOMLEFT) [c]
-- relativePoint - point on this widget to position relative to (default: BOTTOMLEFT) [c]
-- layout - table of attributes used to configure the icon group
-- sfx - animations, color changes, start and expiration sounds
-- depth - display depth for this view [c]
-- scale - factor to scale view's display size [d]
-- alpha - alpha for the icon group [d]
-- hidden - hide the view [d]

-- Layout attributes:
-- defaults - table with default settings, including those for each class of widget [c]
-- button - widget template for an action button [c]
-- timer - widget template for timer text field [c]
-- count - widget template for count text field [c]
-- backdrop - widget template for backdrop panel that dynamically scales with active icons [c]
-- orientation - "horizontal" or "vertical" (default: "horizontal") [c]
-- direction - true for "left" or "up", false for "right" or "down" [c]
-- limit - maximum number of icons to show (default: 10) [c]
-- spacingX - horizontal distance between icons (default: 0) [c]
-- spacingY - vertical distance between icons (default: 0) [c]
-- wrap - how many icons to show before wrapping to next row or column (default: 0, i.e., don't wrap) [c]
-- wrapDirection - true for "left" or "up", false for "right" or "down" [c]
-- center - if true then snap visible icons to center at anchor [d]

local View_Class = {}
local EF = Puffin.ErrorFormat
local DB, LSM
local recycledViews = {}
local viewNumber = 0

-- Return an initialized instance of a view
function View_Class.CreateView()
	local n, view = next(recycledViews) -- get any available recycled widget
	if n then
		recycledViews[n] = nil
	else
		view = {}
		viewNumber = viewNumber + 1
		view._widgets = {}
	end
	return view
end

-- Release a view when no longer needed, recycle after clearing attributes and hiding widgets associated with it
function View_Class.ReleaseView(view)
	for k, w in pairs(view._widgets) do w._class.Release(w); view._widgets[k] = nil end -- release widgets used by this view
	for k in pairs(view) do if (k ~= "_widgets") then view[k] = nil end end -- clear attributes for this view
	table.insert(recycledViews, view)
end

-- Check an attribute to see if it refers to special media
function View_Class.CheckMedia(v)
	if type(v) == "table" and v[1] == "#LSM" then v = LSM:Fetch(v[2], v[3]) end -- check for LibSharedMedia references
	return v
end

-- Create widget based on its descriptor
function View_Class.CreateWidget(w)
	local class = Puffin.Component(w.class)
	assert(class, EF("View_Class", "unrecognized class"))
	local t = class.Create()
	assert(t, EF("View_Class", "unable to create widget"))
	t._class = class; t.width = w.width; t.height = w.height
	return t
end

-- Configure a widget based on its group and descriptor settings
function View_Class.ConfigureWidget(grp, t, w, offsetDepth, noDimensions)
	local x = (t.offsetX or 0) + (w.offsetX or 0) + ((w.relativeX or 0) * (grp.width or 0))
	local y = (t.offsetY or 0) + (w.offsetY or 0) + ((w.relativeY or 0) * (grp.height or 0))
	if t.parent then t._class.SetParent(t, t.parent) end
	t._class.SetPosition(t, t.anchorPoint or w.anchorPoint or grp.anchorPoint or "BOTTOMLEFT", t.anchor or w.anchor or nil,
		t.relativePoint or w.relativePoint or grp.relativePoint or "BOTTOMLEFT", x, y, t.anchorRegion or w.anchorRegion or nil)
	if not noDimensions then
		t.width = (t.width or w.width or 0) + ((w.relativeWidth or 0) * (grp.width or 0))
		t.height = (t.height or w.height or 0) + ((w.relativeHeight or 0) * (grp.height or 0))
		t.iconSize = w.iconSize or grp.iconSize or 0
	end
	t.scale = w.scale or 1; t.alpha = w.alpha or 1; t.depth = (w.depth or 0) + offsetDepth
	if w.attributes then for i, v in pairs(w.attributes) do t[i] = View_Class.CheckMedia(v) end end -- widget specific attributes
	local defaults = DB.GetRecord("widget_defaults", grp.defaults)
	if defaults and defaults[w.class] then -- default attributes for this widget class
		for i, v in pairs(defaults[w.class]) do if t[i] == nil then t[i] = View_Class.CheckMedia(v) end end
	end
	t._class.Configure(t)
end

-- Create a backdrop for a view
function View_Class.CreateBackdrop(w)
	if w then return View_Class.CreateWidget(w) end -- create backdrop widget using a descriptor, presumably for a panel widget
	return nil
end

-- Configure a backdrop for a view
function View_Class.ConfigureBackdrop(grp, t, w, offsetDepth)
	View_Class.ConfigureWidget(grp, t, w, offsetDepth, true) -- override width/height setting
end

function View_Class.OnEnable()
	DB = Puffin.Component("Data") -- link to data management functions
	assert(DB, EF("View_Class", "Data component not found"))
	LSM = LibStub("LibSharedMedia-3.0")
	assert(LSM, EF("View_Class", "LibSharedMedia not found"))
end

Puffin.RegisterComponent("View_Class", View_Class, "VIEW", 0)
