-- Puffin: A Tojaso FowlPlay Project
-- Widget to show text in a variety of formats (including countdown time)

-- Custom attributes:
-- font - font to use for the text (default: ChatFontNormal) [c]
-- fsize - font size for the text (default: 10) [c]
-- shadow - enable a shadow behind the text (default: false) [c]
-- outline - enable font "OUTLINE" flag (default: false) [c]
-- thick - enable font "THICKOUTLINE" flag (default: false) [c]
-- mono - enable font "MONOCHROME" flag (default: false) [c]
-- wrap - enable wrap of text to multiple lines (default: false) [c]
-- justifyH - horizontal text alignment (default: "CENTER") [c]
-- justifyV - vertical text alignment (default: "MIDDLE") [c]
-- textFormat - type of text to display (default: "text") [c]
-- text - string to display when textFormat is "text" (default: "") [d]
-- timeExpire - expiration time when textFormat is "time" [d]
-- timeMax - maximum value of time to display when textFormat is "time" [d]
-- timeLeft - override time computation when textFormat is "time" [d]
-- timeFormat - option format for converting time to text when textFormat is "time" (default: most compact) [d]
-- timeSpaces - option to remove spaces in time formats (default: false) [d]
-- timeCase - option to convert time text characters to lowercase (default: false) [d]
-- color - color to apply to text (default: none) [d]

local Widget_Text = {}
local EF = Puffin.ErrorFormat
local recycledText = {}
local activeText = {}
local textNumber = 0

-- Convert a time value into a compact text string
local TimeFormatOptions = { { 1, 1, 1, 1, 1 }, { 1, 1, 1, 3, 5 }, { 1, 1, 1, 3, 4 },
						{ 2, 3, 1, 2, 3 }, { 2, 3, 1, 2, 2 }, { 2, 3, 1, 3, 4 }, { 2, 3, 1, 3, 5 },
						{ 2, 2, 2, 2, 3 }, { 2, 2, 2, 2, 2 }, { 2, 2, 2, 2, 4 }, { 2, 2, 2, 3, 4 }, { 2, 2, 2, 3, 5 },
						{ 2, 3, 2, 2, 3 }, { 2, 3, 2, 2, 2 }, { 2, 3, 2, 2, 4 }, { 2, 3, 2, 3, 4 }, { 2, 3, 2, 3, 5 },
						{ 2, 3, 3, 2, 3 }, { 2, 3, 3, 2, 2 }, { 2, 3, 3, 2, 4 }, { 2, 3, 3, 3, 4 }, { 2, 3, 3, 3, 5 },
						{ 3, 3, 3, 2, 3 }, { 3, 3, 3, 3, 5 },
						{ 4, 3, 1, 2, 3 }, { 4, 3, 1, 2, 2 }, { 4, 3, 1, 3, 4 }, { 4, 3, 1, 3, 5 },
						{ 5, 1, 1, 2, 3 }, { 5, 1, 1, 2, 2 }, { 5, 1, 1, 3, 4 }, { 5, 1, 1, 3, 5 },
}

local function FormatTime(t, timeFormat, timeSpaces, timeCase)
	if not timeFormat or (timeFormat > #TimeFormatOptions) then timeFormat = 24 end -- default to most compact
	local opt = TimeFormatOptions[timeFormat]
	local h, m, hplus, mplus, s, ts, f
	local o1, o2, o3, o4, o5 = opt[1], opt[2], opt[3], opt[4], opt[5]
	h = math.floor(t / 3600); m = math.floor((t - (h * 3600)) / 60); s = math.floor(t - (h * 3600) - (m * 60))
	hplus = math.floor((t + 3599.99) / 3600); mplus = math.floor((t - (h * 3600) + 59.99) / 60) -- provides compatibility with tooltips
	ts = math.floor(t * 10) / 10 -- truncated to a tenth second
	if t >= 3600 then
		if o1 == 1 then f = string.format("%.0f:%02.0f:%02.0f", h, m, s) elseif o1 == 2 then f = string.format("%.0fh %.0fm", h, m)
			elseif o1 == 3 then f = string.format("%.0fh", hplus) elseif o1 == 4 then f = string.format("%.0fh %.0f", h, m)
			else f = string.format("%.0f:%02.0f", h, m) end
	elseif t >= 120 then
		if o2 == 1 then f = string.format("%.0f:%02.0f", m, s) elseif o2 == 2 then f = string.format("%.0fm %.0fs", m, s)
			else f = string.format("%.0fm", mplus) end
	elseif t >= 60 then
		if o3 == 1 then f = string.format("%.0f:%02.0f", m, s) elseif o3 == 2 then f = string.format("%.0fm %.0fs", m, s)
			else f = string.format("%.0fm", mplus) end
	elseif t >= 10 then
		if o4 == 1 then f = string.format(":%02.0f", s) elseif o4 == 2 then f = string.format("%.0fs", s)
			else f = string.format("%.0f", s) end
	else
		if o5 == 1 then f = string.format(":%02.0f", s) elseif o5 == 2 then f = string.format("%.1fs", ts)
			elseif o5 == 3 then f = string.format("%.0fs", s) elseif o5 == 4 then f = string.format("%.1f", ts)
			else f = string.format("%.0f", s) end
	end
	if not timeSpaces then f = string.gsub(f, " ", "") end
	if timeCase then f = string.upper(f) end
	return f
end

-- Convert time to string for this text widget
local function ConvertTimeToText(w)
	local s = ""
	local timeLeft = w.timeLeft
	if not timeLeft and w.timeExpire then timeLeft = w.timeExpire - GetTime() end
	if timeLeft and timeLeft >= 0 and (not w.timeMax or timeLeft < w.timeMax) then -- check if expiration time reached or timeLeft is longer than maximum limit
		s = FormatTime(timeLeft, w.timeFormat, w.timeSpaces, w.timeCase)
	end
	return s
end

-- Create a new text widget object
function Widget_Text.Create()
	local n, text = next(recycledText) -- get any available recycled widget
	if n then
		recycledText[n] = nil
	else
		text = {}
		textNumber = textNumber + 1
		local t = {}; text._display = t -- table of display objects preserved when widget is recycled
		local frameName = "Puffin_Text" .. tostring(textNumber) -- generate unique name for text's frame
		t.container = CreateFrame("Frame", frameName .. "Container", UIParent)
		t.container:SetFrameStrata("MEDIUM")
		t.fontstring = t.container:CreateFontString(nil, "ARTWORK", ChatFontNormal)
	end
	table.insert(activeText, text) -- keep track of widgets so can animate if necessary
	return text
end

-- Release text when no longer needed, recycle object after clearing attributes and hiding display objects associated with it
function Widget_Text:Release()
	for k in pairs(self) do if k ~= "_display" then self[k] = nil end end
	self._display.fontstring:Hide(); self._display.container:Hide()
	for k, v in pairs(activeText) do if v == self then activeText[k] = nil end end
	table.insert(recycledText, self)
end

-- Set display position for the widget relative to another widget or the overall screen
function Widget_Text:SetPosition(anchor, anchorWidget, relative, offsetX, offsetY, region)
	local t, w = self._display, UIParent
	if anchorWidget then
		if anchorWidget._display then w = region and anchorWidget._display[region] or anchorWidget._display.container else w = anchorWidget end
	end
	t.container:ClearAllPoints(); t.container:SetPoint(anchor, w, relative, offsetX, offsetY)
end

-- Set the parent widget or frame (used for managing display heirarchy)
function Widget_Text:SetParent(widget)
	local t, w = self._display, widget._display
	if w and w.container then w = w.container else w = widget end
	if not w then w = UIParent end
	t.container:SetParent(w)
	if not self.depth then t.container:SetFrameLevel(w:GetFrameLevel() + 1) end
end

-- Update depth for the widget
function Widget_Text:SetDepth(depth)
	local t = self._display; if not t then return end
	local b = t.container; if not b then return end
	self.depth = depth
	if self.depth then b:SetFrameLevel(self.depth) end
end

-- Return encoded text flags for outline, thick, mono attributes
local function TextFlags(outline, thick, mono)
	local t = nil
	if mono then
		if outline then if thick then t = "MONOCHROME,OUTLINE,THICKOUTLINE" else t = "MONOCHROME,OUTLINE" end
		else if thick then t = "MONOCHROME,THICKOUTLINE" else t = "MONOCHROME" end end
	else
		if outline then if thick then t = "OUTLINE,THICKOUTLINE" else t = "OUTLINE" end
		else if thick then t = "THICKOUTLINE" end end
	end
	return t
end

-- Configure all the text's display objects based on current attribute values (or defaults)
function Widget_Text:Configure()
	assert(self.width and self.width > 0, EF("Widget_Text", "undefined width"))
	assert(self.height and self.height > 0, EF("Widget_Text", "undefined height"))
	local tc, fs = self._display.container, self._display.fontstring	
	tc:SetSize(self.width, self.height); tc:SetScale(self.scale or 1)
	fs:SetJustifyH(self.justifyH or "CENTER"); fs:SetJustifyV(self.justifyV or "MIDDLE"); fs:SetWordWrap(self.wrap)
	fs:SetShadowColor(0, 0, 0, self.shadow and 1 or 0); fs:SetShadowOffset(1, -1); fs:SetFontObject(ChatFontNormal)
	fs:SetSize(self.width, self.height); fs:ClearAllPoints(); fs:SetPoint("CENTER", tc, "CENTER")
	-- tc.tex = tc:CreateTexture(nil, "ARTWORK"); tc.tex:SetAllPoints(); tc.tex:SetTexture(1,0,0,0.5); tc.tex:Show() -- debug
	fs:Hide(); tc:Hide()
end

-- Update just the text and color displayed by this widget
function Widget_Text:Update()
	local t = self._display; if not t then return end
	local tc, fs = t.container, t.fontstring
	if self.hidden or (self.condition == false) then
		fs:Hide(); tc:Hide()
	else
		local d = self.depth; if d then tc:SetFrameLevel(d) end
		local c = self.color; if c then fs:SetTextColor(c.r, c.g, c.b) else fs:SetTextColor(0, 0, 0) end
		if self.font then fs:SetFont(self.font, self.fsize or 10, TextFlags(self.outline, self.thick, self.mono)) end
		if self.textFormat == "time" then self.text = ConvertTimeToText(self) end
		fs:SetText(self.text); tc:SetAlpha(self.alpha or 1); fs:Show(); tc:Show()
		self.textWidth = fs:GetStringWidth() -- save the actual displayed width of the text
	end
end

-- Update the text's animated display objects (i.e., time countdown)
function Widget_Text:Animate()
	if not self.hidden and (self.condition ~= false) and (self.textFormat == "time") then
		local fs = self._display.fontstring
		if fs then self.text = ConvertTimeToText(self); fs:SetText(self.text) end
	end
end

-- Run animations for all active widgets
function Widget_Text:OnRefresh() for _, w in pairs(activeText) do Widget_Text.Animate(w) end end

-- Register this widget as a component so it can be found by views
Puffin.RegisterComponent("Widget_Text", Widget_Text, "WIDGET", 0)
